import logging
import threading
from getpass import getpass

from dvc_objects.fs.base import FileSystem
from funcy import cached_property, memoize, wrap_prop, wrap_with

logger = logging.getLogger(__name__)


@wrap_with(threading.Lock())
@memoize
def ask_password(host, user):
    return getpass(f"Enter a password for host '{host}' user '{user}':\n")


class WebDAVFileSystem(FileSystem):  # pylint:disable=abstract-method
    protocol = "webdav"
    root_marker = ""
    REQUIRES = {"webdav4": "webdav4"}
    PARAM_CHECKSUM = "etag"

    def __init__(self, **config):
        super().__init__(**config)

        cert_path = config.get("cert_path", None)
        key_path = config.get("key_path", None)
        self.prefix = config.get("prefix", "")
        self.fs_args.update(
            {
                "base_url": config["url"],
                "cert": cert_path if not key_path else (cert_path, key_path),
                "verify": config.get("ssl_verify", True),
                "timeout": config.get("timeout", 30),
            }
        )

    def unstrip_protocol(self, path: str) -> str:
        return self.fs_args["base_url"] + "/" + path

    @staticmethod
    def _get_kwargs_from_urls(urlpath):
        from urllib.parse import urlparse, urlunparse

        parsed = urlparse(urlpath)
        scheme = parsed.scheme.replace("webdav", "http")
        return {
            "prefix": parsed.path.rstrip("/"),
            "host": urlunparse((scheme, parsed.netloc, "", None, None, None)),
            "url": urlunparse(
                (
                    scheme,
                    parsed.netloc,
                    parsed.path.rstrip("/"),
                    None,
                    None,
                    None,
                )
            ),
            "user": parsed.username,
        }

    def _prepare_credentials(self, **config):
        user = config.get("user", None)
        password = config.get("password", None)
        headers = {}
        auth = None
        if token := config.get("token"):
            headers.update({"Authorization": f"Bearer {token}"})
        elif user:
            if not password and config.get("ask_password"):
                password = ask_password(config["host"], user)
            auth = (user, password)
        elif custom_auth_header := config.get("custom_auth_header"):
            if not password and config.get("ask_password"):
                password = ask_password(config["host"], custom_auth_header)
            headers.update({custom_auth_header: password})

        return {"headers": headers, "auth": auth}

    @wrap_prop(threading.Lock())
    @cached_property
    def fs(self):
        from webdav4.fsspec import WebdavFileSystem

        return WebdavFileSystem(**self.fs_args)

    def upload_fobj(self, fobj, to_info, **kwargs):
        size = kwargs.get("size")
        # using upload_fileobj to directly upload fileobj rather than buffering
        # and using overwrite=True to avoid check for an extra exists call,
        # as caller should ensure that the file does not exist beforehand.
        return self.fs.upload_fileobj(fobj, to_info, overwrite=True, size=size)


class WebDAVSFileSystem(WebDAVFileSystem):  # pylint:disable=abstract-method
    protocol = "webdavs"



from dvc.testing.cloud import Cloud
from dvc.testing.path_info import WebDAVURLInfo
from funcy import cached_property, first

AUTH = {"user1": {"password": "password1"}}
BASE_PATH = "/dav/files/user1"


class Webdav(Cloud, WebDAVURLInfo):
    @staticmethod
    def get_url(port):  # pylint: disable=arguments-differ
        return f"webdav://localhost:{port}{BASE_PATH}"

    @property
    def config(self):
        user, secrets = first(AUTH.items())
        url = f"webdav://{self.netloc}{self._spath}"
        return {"url": url, "user": user, **secrets}

    @cached_property
    def client(self):
        from webdav4.client import Client

        user, secrets = first(AUTH.items())
        return Client(
            self.replace(path=BASE_PATH).url, auth=(user, secrets["password"])
        )

    def mkdir(self, mode=0o777, parents=False, exist_ok=False):
        assert mode == 0o777
        parent_dirs = list(reversed(self.parents))[1:] if parents else []
        for d in parent_dirs + [self]:
            path = d.path  # pylint: disable=no-member
            if not self.client.exists(path):
                self.client.mkdir(path)

    def write_bytes(self, contents):
        from io import BytesIO

        self.client.upload_fileobj(BytesIO(contents), self.path)

    @property
    def fs_path(self):
        return self.path.lstrip("/")

    def exists(self):
        return self.client.exists(self.path)

    def is_dir(self):
        return self.client.isdir(self.path)

    def is_file(self):
        return self.client.isfile(self.path)

    def read_bytes(self):
        with self.client.open(self.path, mode="rb") as f:
            return f.read()



import os

import pytest
from cheroot import wsgi
from wsgidav.wsgidav_app import WsgiDAVApp

from .cloud import AUTH, BASE_PATH, Webdav

# pylint: disable=redefined-outer-name


@pytest.fixture
def webdav_server(tmp_path_factory):
    host, port = "localhost", 0
    directory = os.fspath(tmp_path_factory.mktemp("http"))
    dirmap = {BASE_PATH: directory}

    app = WsgiDAVApp(
        {
            "host": host,
            "port": port,
            "provider_mapping": dirmap,
            "simple_dc": {"user_mapping": {"*": AUTH}},
        }
    )

    server = wsgi.Server(bind_addr=(host, port), wsgi_app=app)
    with server._run_in_thread():  # pylint: disable=protected-access
        yield server


@pytest.fixture
def make_webdav(webdav_server):
    def _make_webdav():
        url = Webdav.get_url(webdav_server.bind_addr[1])
        return Webdav(url)

    return _make_webdav


@pytest.fixture
def webdav(make_webdav):
    return make_webdav()
